/*
    Copyright (C) 1997-2001  Ulric Eriksson <ulric@siag.nu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA.
*/

/*
 * fonts.c
 *
 * Support for figuring out richtext metrics and displaying rich text.
 * Includes support for T1lib.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <X11/cursorfont.h>
#include "../config.h"
#ifdef HAVE_LIB_T1X
#include <t1lib.h>
#include <t1libx.h>
#endif

#include "../common/common.h"
#include <Mowitz/Mowitz.h>
#include "../common/oldfonts.h"
#include "../common/fonts.h"
#include "../common/traceme.h"
#include "xfonts.h"

#define has_t1_font(i) (font_table[(i)].id != -1)

static Display *dpy;

siag_color color_table[1000];
siag_font font_table[256];
static int ncolor, nfont;
static int scale = 1000;

static Display *t1_dpy;
static Screen *t1_scr;
static int t1_scrn;
static int t1_nfont;
static char **t1_fontname;
static int t1_mode;

static int compar(const void *p, const void *q)
{
	return strcmp(*(const char **)p, *(const char **)q);
}

/* ---
returns a null-terminated, sorted list of all known fonts.
Caller must free
*/

char **font_list(int *n)
{
	int i, j;
	char **list = (char **)MwMalloc((nfontalias+nfontname+1)*sizeof(char *));
	for (i = 0; i < nfontalias; i++)
		list[i] = MwStrdup(fontalias[i].alias);
	for (j = 0; j < nfontname; j++)
		list[i+j] = MwStrdup(fontname_table[j].name);
	list[i+j] = NULL;
	qsort(list, i+j, sizeof(char *), compar);
	*n = i+j;
	return list;
}

/* ---
returns a null-terminated, sorted list of all known colours.
Caller must free
*/

char **color_list(int *n)
{
	int i;
	char **list = (char **)MwMalloc((ncolor+1)*sizeof(char *));
	for (i = 0; i < ncolor; i++)
		list[i] = MwStrdup(color_table[i].name);
	list[i] = NULL;
	qsort(list, i, sizeof(char *), compar);
	*n = i;
	return list;
}

/* ---
The index is into the font_table. The index must have come from
lookup_font, so we know that the font is already loaded.
*/

#define font_struct(i) (font_table[(i)].fs)

int font_ascent(int index)
{
	if (has_t1_font(index)) {
#ifdef HAVE_LIB_T1X
		GLYPH *glyph = T1_SetChar(font_table[index].id, 'd',
				font_table[index].size/10.0, NULL);
		if (glyph) return glyph->metrics.ascent;
#endif
		return font_table[index].size/10;	/* BOGUS */
	} else {
		return font_struct(index)->max_bounds.ascent;
	}
}

int font_descent(int index)
{
	if (has_t1_font(index)) {
#ifdef HAVE_LIB_T1X
		GLYPH *glyph = T1_SetChar(font_table[index].id, 'p',
				font_table[index].size/10.0, NULL);
		if (glyph) return -glyph->metrics.descent;
#endif
		return font_table[index].size/10;	/* BOGUS */
	} else {
		return font_struct(index)->max_bounds.descent;
	}
}

int font_height(int index)
{
#ifdef HAVE_LIB_T1X
	if (has_t1_font(index)) {
		BBox bbox = T1_GetFontBBox(font_table[index].id);
		return (bbox.ury-bbox.lly)*font_table[index].size/10000;
	}
#endif
	return font_ascent(index)+font_descent(index);
}

int font_width(int index, char *s)
{
	if (has_t1_font(index)) {
#ifdef HAVE_LIB_T1X
		return T1_GetStringWidth(font_table[index].id,
			s, strlen(s), 0, FALSE)*font_table[index].size/10000;
#else
		return font_table[index].size/20;
#endif
	} else {
		return XTextWidth(font_struct(index), s, strlen(s));
	}
}

/* ---
index to color_table. The index must come from lookup_color,
so we know that the colour has already been allocated
*/

unsigned long get_color(int index)
{
	return color_table[index].color.pixel;
}

/* ---
Take an index into the format table (i.e. a new-style format code) and
decode it into the components specified by the mask.
*/

void decode_format(int n, long mask, sfmt *fmt)
{
	int font = format_table[n].font;

	if (mask & SFMT_FAMILY) {
		fmt->family = fontname_table[font_table[font].name].name;
		if (!fmt->family) fmt->family = "Helvetica";
	}
	if (mask & SFMT_SIZE)
		fmt->size = font_table[font].size;
	if (mask & SFMT_BOLD)
		fmt->bold = font_table[font].bold;
	if (mask & SFMT_ITALIC)
		fmt->italic = font_table[font].italic;
	if (mask & SFMT_ULINE)
		fmt->uline = format_table[n].uline;
	if (mask & SFMT_STRIKE)
		fmt->strike = format_table[n].strike;
	if (mask & SFMT_FG) {
		fmt->fg = color_table[format_table[n].fg].name;
		if (!fmt->fg) fmt->fg = "black";
	}
	if (mask & SFMT_BG) {
		fmt->bg = color_table[format_table[n].bg].name;
		if (!fmt->bg) fmt->bg = "white";
	}
	if (mask & SFMT_BORDERS)
		fmt->borders = format_table[n].siagfmt & BORDER_MASK;
	if (mask & SFMT_VADJ)
		fmt->vadj = format_table[n].siagfmt & VADJ_MASK;
	if (mask & SFMT_HADJ)
		fmt->hadj = format_table[n].siagfmt & HADJ_MASK;
	if (mask & SFMT_STYLE)
		fmt->style = format_table[n].style;
}

/* ---
Encode the components from fmt specified by the mask into a new-style
format code, allocating colours and fonts as necessary.
*/

int encode_format(long mask, sfmt *fmt)
{
	char *family = "Helvetica";
	int fg = 0, bg = 0;
	int size = 120;
	int bold = 0, italic = 0, uline = 0, strike = 0;
	int siagfmt = 0;
	int style = 0;

	if (mask & SFMT_FAMILY) {
		family = lookup_fontalias(fmt->family);
		if (!family) family = "Helvetica";
	}
	if (mask & SFMT_SIZE)
		size = fmt->size;
	if (mask & SFMT_BOLD)
		bold = fmt->bold;
	if (mask & SFMT_ITALIC)
		italic = fmt->italic;
	if (mask & SFMT_ULINE)
		uline = fmt->uline;
	if (mask & SFMT_STRIKE)
		strike = fmt->strike;
	if (mask & SFMT_FG) {
		fg = lookup_color(fmt->fg);
		if (fg == -1) fg = 0;	/* black */
	}
	if (mask & SFMT_BG) {
		bg = lookup_color(fmt->bg);
		if (bg == -1) bg = 7;	/* white */
	}
	if (mask & SFMT_BORDERS)
		siagfmt |= fmt->borders;
	if (mask & SFMT_VADJ)
		siagfmt |= fmt->vadj;
	if (mask & SFMT_HADJ)
		siagfmt |= fmt->hadj;
	if (mask & SFMT_STYLE)
		style = fmt->style;
	return lookup_format(lookup_font(family, size, bold, italic),
				uline, strike, fg, bg, style, siagfmt);
}

/* ---
Here is the whole song and dance to get a format id for Siag versions >= 3.1.

Start by deciding on font family, point size, weight and slant.
Get an index into the font table by calling lookup_font with
the family name as argument along with the size, weight and slant.
The font is automatically loaded, if necessary.

Next pick foreground and background colours. The colours are loaded
by calling lookup_color with the colour name as argument. An index into
the colour table is returned.

Finally we are ready to call lookup_format. Decide if the character
should be underlined. If the application is Siag, also combine the
cell attributes. Pass the font index, the colour index, the underlining
flag and the Siag attributes to lookup_format. An index into the
format table is returned, and it is this index we can use to replace
the old style formats.

Yes, this is more complex than the old stuff. But it is also in fact
easier to extract individual pieces of the formats, since they are
normal C structures. To my eyes, font_table[format_table[fmt].font].size
is much clearer than (fmt & SIZE_MASK)>>SIZE_SHIFT. And it is less
prone to errors. And it leaves room for future expansion, for example
for overstrike and double underline, or bidirectional printing.
--- */

/* ---
Initialise the X format code.
*/

void init_format(Display *d)
{
	dpy = d;
}


/* ---
returns the index of a structure with the requested color, or -1
   for failure. Pixel value in color_table[i].color.pixel is allocated
*/

int lookup_color(char *name)
{
	int i;
	Colormap cmap;

	if (!name) {
		abort();
	}
	if (!dpy) return 0;

	cmap = DefaultColormap(dpy, DefaultScreen(dpy));

	for (i = 0; i < ncolor; i++) {
		if (!MwStrcasecmp(name, color_table[i].name))
			break;
	}
	if (i == ncolor) return -1;	/* fail */
	if (color_table[i].need_init) {
		color_table[i].color.red = color_table[i].red;
		color_table[i].color.green = color_table[i].green;
		color_table[i].color.blue = color_table[i].blue;
		color_table[i].color.flags = DoRed|DoGreen|DoBlue;
		if (!XAllocColor(dpy, cmap, &color_table[i].color))
			return -1;
		color_table[i].need_init = 0;
	}
	return i;
}

/* ---
for the given font with the given properties, encode X, T1 and PS strings.
example (result in x_name will vary depending on resolution):

	font = index of font with name "Helvetica"
	size = 12
	bold = 0
	italic = 1;

	x_name = "-adobe-helvetica-medium-o-*--*-120-75-75-*-*-iso8859-1"
	t1_name = "Helvetica-Oblique"
	ps_name = "Helvetica-Oblique"

t1_name may be NULL. The caller must then use the X font.
*/

static void encode_font(Display *dpy,
		int font, int size, int bold, int italic, float zoom,
		char *x_name, char **t1_name, char **ps_name, int **afm)
{
	char *x_fmt;
	int i = 0;
	int res_x, res_y;	/* screen resolution */
	res_x = res_y = 72*zoom;

	if (bold) i += 2;
	if (italic) i += 1;

	x_fmt = fontname_table[font].x_name[i];
	if (!x_fmt) x_fmt = fontname_table[0].x_name[0];
	sprintf(x_name, x_fmt, size, res_x, res_y);
	*t1_name = fontname_table[font].t1_name[i];
	*ps_name = fontname_table[font].ps_name[i];
	*afm = fontname_table[font].afm[i];
}

static struct {
	char *name;
	XFontStruct *fs;
} *loaded_fonts = NULL;

XFontStruct *load_cached_font(char *result, char *family, int size,
		int bold, int italic, float zoom)
{
	char *t1_name, *ps_name;
	int i, *afm;
	XFontStruct *fs;

	encode_font(dpy, lookup_fontname(family),
		size, bold, italic, zoom, result, &t1_name, &ps_name, &afm);
	if (loaded_fonts == NULL) {
		loaded_fonts = MwMalloc(1*sizeof *loaded_fonts);
		loaded_fonts[0].name = NULL;
	}
	for (i = 0; loaded_fonts[i].name; i++) {
		if (!strcmp(loaded_fonts[i].name, result))
			return loaded_fonts[i].fs;
	}
	loaded_fonts = MwRealloc(loaded_fonts, (i+1)*sizeof(*loaded_fonts));
	loaded_fonts[i].name = MwStrdup(result);
	fs = XLoadQueryFont(dpy, result);
	if (fs == NULL) {
		fs = XLoadQueryFont(dpy, "fixed");
	}
	loaded_fonts[i++].fs = fs;
	loaded_fonts[i].name = 0;
	return fs;
}

/* ---
return index into font table or -1 for failure. Load font if necessary.
This function is wasteful in that it always loads an X font, even if
a T1 font is found and will be used.
*/

int lookup_font(char *fname, int size, int bold, int italic)
{
	int i, name, *afm;
	char x_name[1000], *t1_name, *ps_name;

	name = lookup_fontname(fname);
	if (name == -1) name = 0;	/* fallback value */

	for (i = 0; i < nfont; i++) {
		if (font_table[i].name == name &&
			font_table[i].size == size &&
			font_table[i].bold == bold &&
			font_table[i].italic == italic)
				return i;	/* already loaded */
	}
	nfont++;
	font_table[i].name = name;
	font_table[i].size = size;
	font_table[i].bold = bold;
	font_table[i].italic = italic;
	encode_font(dpy, name, size, bold, italic, 1.0,
		x_name, &t1_name, &ps_name, &afm);
	font_table[i].x_name = MwStrdup(x_name);
	if (ps_name) font_table[i].ps_name = MwStrdup(ps_name);
	else font_table[i].ps_name = NULL;
	if (t1_name) font_table[i].t1_name = MwStrdup(t1_name);
	else font_table[i].t1_name = NULL;
	font_table[i].afm = afm;
	font_table[i].fs = XLoadQueryFont(dpy, x_name);
	font_table[i].id = t1_getfontid(t1_name);
	if (font_table[i].fs == NULL) {
		font_table[i].fs =
			XLoadQueryFont(dpy, "*helvetica-medium-r*12*");
	}
	if (font_table[i].fs == NULL) {
		fprintf(stderr, "Panic: can't load any fonts!\n");
		exit(EXIT_FAILURE);
	}
	return i;
}

/* ---
stores a colour name along with the red, green and blue values,
   or overwrites an existing colour with the same name.
   The colour is not allocated until it is looked up in lookup_color.
*/

int register_color(char *name,
	unsigned short red, unsigned short green, unsigned short blue)
{
	int i;

	for (i = 0; i < ncolor; i++)
		if (!MwStrcasecmp(name, color_table[i].name)) break;
	if (i == ncolor) {
		ncolor++;
		color_table[i].name = MwStrdup(name);
	}
	color_table[i].red = red;
	color_table[i].green = green;
	color_table[i].blue = blue;
	color_table[i].need_init = 1;
	return i;
}


/* ---
fmt is an index into fontname_table
*/

char *family2name(int fmt)
{
	return fontname_table[fmt].name;
}

/* ---
fmt is index into font_table
*/

char *size2name(int fmt)
{
	static char b[80];

	sprintf(b, "%d", font_table[fmt].size/10);
	return b;
}

/* ---
fmt is index into color_table
*/

char *color2name(int fmt)
{
	return color_table[fmt].name;
}

/* ---
Postscript font name
*/

char *ps_fontname(int fmt)
{
	char *p = font_table[format_table[fmt].font].ps_name;
	if (p == NULL) p = font_table[format_table[fmt].font].t1_name;
	if (p == NULL) p = font_table[0].ps_name;
	return p;
}

/* ---
Postscript font size
*/

int ps_fontsize(int fmt)
{
	return font_table[format_table[fmt].font].size / 10;
}

static int lastfont;

/* ---
Define Latin-1 fonts

Improved: only makes the fonts that are actually used.

This is a mess, but here is a strategy I believe in. Go through the entire
format table, looking for font, italic and bold information. For each
format, call a function that figures out whether to use ps or t1 fonts,
encodes it, uploads the font if necessary and adds it to a list of "done"
fonts. Fonts that have already been done are not repeated even if they are
found again in the format table.

Because of the way the font system works (the same font may be used for
regular and bold, for example), it is necessary to encode the fonts
before checking if it is already done.
*/

static struct done {
	char *name;
	struct done *next;
} *done_fonts;

static void ps_makefont(FILE *fp, int id, char *ps_name, int pfab,
			char *enc, char *pfb)
{
	struct done *df;

	for (df = done_fonts; df; df = df->next)
		if (!strcmp(df->name, ps_name)) return;

	df = MwMalloc(sizeof(struct done));
	df->name = MwStrdup(ps_name);
	df->next = done_fonts;
	done_fonts = df;

	if (pfab == 0 && pfb != NULL) {
		FILE *fpin;
		char b[1024];

		sprintf(b, "%s/common/fonts/%s", datadir, pfb);
		fpin = fopen(b, "r");
		if (fpin) {
			pfb2ps(fpin, fp);
			fclose(fpin);
		}
	}

#ifdef HAVE_LIB_T1X
	if (pfab) {	/* dump font into ps file */
#if 1
		FILE *fpin;
		char b[1024];

		if (id == -1) return;
		sprintf(b, "%s/common/fonts/%s", datadir,
			T1_GetFontFileName(id));
		fpin = fopen(b, "r");
		if (fpin) {
			pfb2ps(fpin, fp);
			fclose(fpin);
		}
#else
		char b[1024], cmd[10000];
		FILE *ffp;
		if (id == -1) return;
		sprintf(cmd, "%s/common/readpfa %s %s", datadir,
			T1_GetFileSearchPath(T1_PFAB_PATH),
			T1_GetFontFileName(id));
		ffp = popen(cmd, "r");
		if (ffp == NULL) {
			fprintf(stderr, "Couldn't read from %s\n", cmd);
			return;
		}
		while (fgets(b, sizeof b, ffp))
			fputs(b, fp);
		pclose(ffp);
		/* and presto, we have a font */
#endif
	}
#endif
	if (!enc) return;	/* don't bother */

	fprintf(fp, "/%s findfont\n", ps_name);

	fprintf(fp, "dup length dict begin\n");
	fprintf(fp, "  {1 index /FID ne {def} {pop pop} ifelse} forall\n");
	fprintf(fp, "  /Encoding %sEncoding def\n", enc);
	fprintf(fp, "  currentdict\n");
	fprintf(fp, "end\n");
	fprintf(fp, "/%s-%s exch definefont pop\n", ps_name, enc);
}

void ps_makefonts(FILE *fp)
{
	int i, j;
	struct done *df;

	done_fonts = NULL;

	for (i = 0; i < nformat; i++) {
		char *ps_name, *enc;
		int fi = format_table[i].font;	/* index to font_table */
		int id = font_table[fi].id;
		int bold = font_table[fi].bold;
		int italic = font_table[fi].italic;
		int name = font_table[fi].name;
		j = 0;
		if (bold) j += 2;
		if (italic) j += 1;

		if (fontname_table[name].iso8859_1) enc = "ISOLatin1";
		else enc = NULL;

		ps_name = fontname_table[name].ps_name[j];
		if (ps_name) {
			ps_makefont(fp, id, ps_name, 0, enc,
				fontname_table[name].pfb[j]);
		} else {
			ps_name = fontname_table[name].t1_name[j];
			if (!ps_name) continue;		/* give up */
			ps_makefont(fp, id, ps_name, 1, enc,
				fontname_table[name].pfb[j]);
		}
	}
	while (done_fonts) {
		df = done_fonts->next;
		MwFree(done_fonts->name);
		MwFree(done_fonts);
		done_fonts = df;
	}
}

static siag_color lastcolor;

void ps_set_color(FILE *fp, int red, int green, int blue)
{
	if (red == lastcolor.red
	    && green == lastcolor.green
	    && blue == lastcolor.blue) return;
	lastcolor.red = red;
	lastcolor.green = green;
	lastcolor.blue = blue;

	if (red == -1) return;		/* starter mark */

	fprintf(fp, "%f %f %f setrgbcolor\n",
		(float)red / 65535.0,
		(float)green / 65535.0,
		(float)blue / 65535.0);
}

/* ---
Set the font
*/

void ps_set_font(FILE *fp, int newfont)
{
	int name;
	siag_color newcolor;

	/* always set color */
	if (newfont == -1) {
		ps_set_color(fp, 0, 0, 0);
	} else {
		newcolor = color_table[format_table[newfont].fg];
		ps_set_color(fp, newcolor.red, newcolor.green, newcolor.blue);
	}

	if (newfont == lastfont) {
		return;
	}
	lastfont = newfont;
	if (newfont == -1) {
		return;	/* starter mark */
	}

	name = font_table[format_table[newfont].font].name;

	fprintf(fp, "/%s%s findfont\n",
		ps_fontname(newfont),
		fontname_table[name].iso8859_1?"-ISOLatin1":"");
	fprintf(fp, "%d scalefont\n", ps_fontsize(newfont));
	fprintf(fp, "setfont\n");
}


/* ---
   Code to handle a combination of X and T1 fonts. Currently pretty
   suboptimal: makes several passes over strings, doesn't use kerning
   or ligatures.

   000425: added zoom.
--- */

/* X code borrowed from Richchar.c */

float rc_width(rich_char c)
{
	int w;
	int index = format_table[c.fmt].font;
	int n = font_table[index].name;		/* index into fontname_table */
	int x = (font_table[index].bold?2:0)+(font_table[index].italic?1:0);
	int *a = fontname_table[n].afm[x];	/* widths */

	if (a) {
		w = a[c.c];
		if (w == 0) {
			/*printf("0 for %d (%c)\n", c.c, c.c);*/
			w = a[0];
		}
		return (float)w*font_table[index].size/10000;
	} else {
		;
	}

	if (has_t1_font(index)) {
#ifdef HAVE_LIB_T1X
		return (float)T1_GetCharWidth(font_table[index].id, c.c)
			*font_table[index].size/10000;
#else
		return (float)font_table[index].size/10;	/* BOGUS */
#endif
	} else {
		w = XTextWidth(font_struct(index), &c.c, 1);
	}
	return w;
}

float rc_strwidth(rich_char *p, int len)
{
	int i;
	float w = 0;

	if (len == -1) len = rc_strlen(p);

	for (i = 0; i < len; i++) {
		w += rc_width(p[i]);
	}
	return w;
}

int rc_height(rich_char c)
{
	int index = format_table[c.fmt].font;

	return font_height(index);
}

int rc_strheight(rich_char *p, int len)
{
	int i, h = 0;

	if (len == -1) len = rc_strlen(p);

	for (i = 0; i < len; i++) {
		int h1 = rc_height(p[i]);
		if (h1 > h) h = h1;
	}
	return h;
}

void rc_strgeom(rich_char *p, int len, int *width, int *height)
{
	*width = rc_strwidth(p, len);
	*height = rc_strheight(p, len);
}

/*
x and y are the unzoomed coordinates, i.e. if x=10 and zoom=2.0,
the text will be drawn at x=20.
*/
int rc_draw(Drawable d, GC gc, int xo, int yo,
		int x, int y, rich_char c, int opaque, float zoom)
{
	static int format_index = -1;
	static sfmt fmt;
	static Font font;
	static float oldzoom = -1;
	XFontStruct *font_struct;
	char xfont[1000];
	int index, color_index;
	int vadj, width;
	unsigned long color;

	if (c.fmt != format_index || zoom != oldzoom) {
		format_index = c.fmt;
		oldzoom = zoom;
		decode_format(format_index,
			SFMT_FAMILY | SFMT_SIZE | SFMT_BOLD | SFMT_ITALIC
			 | SFMT_ULINE | SFMT_FG | SFMT_BG | SFMT_VADJ
			 | SFMT_HADJ, &fmt);
		font_struct = load_cached_font(xfont, fmt.family,
				fmt.size, fmt.bold, fmt.italic, zoom);
		font = font_struct->fid;
		color_index = format_table[c.fmt].fg;
		color = get_color(color_index);
		XSetForeground(t1_dpy, gc, color);

	}
	index = format_table[c.fmt].font;
	switch (format_table[c.fmt].siagfmt & VADJ_MASK) {
	case VADJ_TOP:
		vadj = -6;
		break;
	case VADJ_BOTTOM:
		vadj = 6;
		break;
	default:
		vadj = 0;
	}

	if (has_t1_font(index)) {
#ifdef HAVE_LIB_T1X
		T1_SetCharX(d, gc, t1_mode, zoom*x, zoom*(y+vadj),
			font_table[index].id, c.c,
			zoom*font_table[index].size/10.0, NULL);
if (opaque) fprintf(stderr, "Warning: can't draw opaque\n");
#endif
	} else {
		XSetFont(t1_dpy, gc, font);
		if (opaque) XDrawImageString(t1_dpy, d, gc,
				xo+zoom*x, yo+zoom*(y+vadj), &c.c, 1);
		else XDrawString(t1_dpy, d, gc,
				xo+zoom*x, yo+zoom*(y+vadj), &c.c, 1);
	}
	if (format_table[c.fmt].uline) {
		width = rc_width(c);
		XDrawLine(t1_dpy, d, gc,
			xo+zoom*x, yo+zoom*(y+1),
			xo+zoom*(x+width), yo+zoom*(y+1));
	}
	if (format_table[c.fmt].strike) {
		width = rc_width(c);
		XDrawLine(t1_dpy, d, gc,
			xo+zoom*x, yo+zoom*(y-4),
			xo+zoom*(x+width), yo+zoom*(y-4));
	}
	return 0;
}


int rc_strdraw(Drawable d, GC gc, int xo, int yo,
		int x, int y, rich_char *p, int len, float zoom)
{
	int i;
	float x0 = x;

	if (len == -1) len = rc_strlen(p);

	for (i = 0; i < len; i++) {
		rc_draw(d, gc, xo, yo, x0, y, p[i], 0, zoom);
		x0 += rc_width(p[i]);
	}
	return 0;
}

int t1_getfontid(char *name)
{
	int i;

	if (!name) return -1;
	for (i = 0; i < t1_nfont; i++) {
		if (t1_fontname[i] && !strcmp(t1_fontname[i], name))
			return i;
	}
	return -1;
}

char *t1_getfontname(int id)
{
	if (id < 0 || id >= t1_nfont) return NULL;
	return t1_fontname[id];
}

int t1_init(Widget w)
{
#ifdef HAVE_LIB_T1X
	int i;
	char **enc;
#endif

	t1_dpy = XtDisplay(w);
	t1_scr = XtScreen(w);
	t1_scrn = DefaultScreen(t1_dpy);
	t1_mode = 0;

#ifdef HAVE_LIB_T1X
	if (!getenv("T1LIB_CONFIG")) {
		char dir[1024];
		sprintf(dir, "T1LIB_CONFIG=%s/common/t1lib.config", datadir);
		putenv(dir);
	}

	T1_SetBitmapPad(16);

        if ((T1_InitLib(NO_LOGFILE) == NULL)) {
                fprintf(stderr, "Initialization of t1lib failed\n");
                return -1;
        }

        t1_nfont = T1_Get_no_fonts();

        T1_SetX11Params(t1_dpy, DefaultVisual(t1_dpy, t1_scrn),
                DefaultDepth(t1_dpy, t1_scrn),
                DefaultColormap(t1_dpy, t1_scrn));

        enc = T1_LoadEncoding("IsoLatin1.enc");
        if (enc == NULL) {
                fprintf(stderr, "Couldn't load encoding\n");
                exit(EXIT_FAILURE);
        } else {
                T1_SetDefaultEncoding(enc);
        }

        t1_fontname = MwMalloc(t1_nfont*sizeof *t1_fontname);
        for (i = 0; i < t1_nfont; i++) {
		char *p;
		T1_LoadFont(i);
		p = T1_GetFontName(i);
		T1_DeleteFont(i);
		if (p == NULL) t1_fontname[i] = NULL;
                else t1_fontname[i] = MwStrdup(p);
        }
#endif	/* HAVE_LIB_T1X */
        return 0;
}

int t1_exit(void)
{
#ifdef HAVE_LIB_T1X
        T1_CloseLib();
#endif
	return 0;
}

/* BOGOSITY */
int scale_set(int s)
{
	int old = scale;
	scale = s;
	/* and now we should really delete all loaded fonts,
	   redraw stuff and so on. But not yet.
	*/
	return old;
}

